class GlobalExceptionMap:
    _exception_map = {}

    @classmethod
    def set(cls, exception_map):
        cls._exception_map = exception_map if exception_map is not None else {}

    @classmethod
    def get(cls):
        return dict(cls._exception_map)
